"""Top-level package for sarwaveifrproc."""

__author__ = """Antoine Grouazel"""
__email__ = 'antoine.grouazel@ifremer.fr'

try:
    from importlib import metadata
except ImportError: # for Python<3.8
    import importlib_metadata as metadata
__version__ = metadata.version('sarwaveifrproc')



import sarwaveifrproc
import tensorflow as tf
import datatree as dtt
import numpy as np
import glob
import logging
import yaml 
import pickle
import re
import os
from datetime import datetime
from sarwaveifrproc.l2_wave import generate_l2_wave_product

def get_safe_date(safe):
    """
    Extracts and returns the date and time from a given SAFE directory name.

    Parameters:
        safe (str): SAFE directory name.

    Returns:
        datetime: A datetime object representing the extracted date and time.
    """
    date_str = safe.split('_')[5]
    date = datetime.strptime(date_str, '%Y%m%dT%H%M%S')
    return date

def get_output_safe(l1x_safe, root_savepath, tail='E00'):
    """
    Generates the final SAFE filename with specified modifications.

    Parameters:
        l1x_safe (str): The input SAFE path.
        root_savepath (str): The root path for saving the output.
        tail (str): The tail to append to the filename. Defaults to 'E00'.

    Returns:
        str: The output savepath.
        
    Raises:
        ValueError: If the input SAFE does not meet the filename requirements.
    """
    l1x_safe = l1x_safe.rstrip('/') # remove trailing slash
    safe = l1x_safe.split(os.sep)[-1]
    final_safe = safe
    
    if all((pattern in safe) for pattern in ['XSP_', '1SDV', '.SAFE']):
        
        date = get_safe_date(safe)

        final_safe = final_safe.replace('XSP_', 'WAV_')
        final_safe = final_safe.replace('1SDV', '2SDV')
        regexA = re.compile("A[0-9]{2}.SAFE")
        regexB = re.compile("B[0-9]{2}.SAFE")
        if  re.search(regexA, final_safe.split('_')[-1]) or re.search(regexB, final_safe.split('_')[-1]):
            print('match regexp')
            final_safe = final_safe.replace(final_safe.split('_')[-1], f'{tail.upper()}.SAFE')
        else:
            print('no slug existing-> just add the product ID')
            final_safe = final_safe.replace('.SAFE', f'_{tail.upper()}.SAFE')

        
        output_safe = os.path.join(root_savepath, date.strftime('%Y'), date.strftime('%j'), final_safe)

        return output_safe
    
    else:
        raise ValueError(f"The input SAFE does not meet the filename requirements.")
    
def get_output_filename(l1x_path, output_safe, tail='e00'):
    """
    Constructs and returns the file path for saving a processed file based on input parameters.
    
    Parameters:
        l1x_path (str): The path to the input file. It can be either a L1B or L1C file.
        root_savepath (str): The root directory where the processed file will be saved.
        tail (str, optional): The tail string to be appended to the filename corresponding to the processing options. Defaults to 'e00'.
        
    Returns:
        str: File path for saving the processed file.
    """
    filename = l1x_path.split(os.sep)[-1]
    filename_exploded = filename.split('-')
    regex_file_number = re.compile("[0-9]{2}[0-6]")
    if filename[0:2]=='l1':
        final_filename = '-'.join([*filename_exploded[:4], 'dv', *filename_exploded[5:-1], f'{tail.lower()}.nc'])
        if re.search(regex_file_number,final_filename.split('-')[9]):
            final_filename = final_filename.replace(final_filename.split('-')[9]+'-','') #remove the -004- giving the number of the file
    else:
        final_filename = '-'.join([*filename_exploded[:3], 'dv', *filename_exploded[4:-1], f'{tail.lower()}.nc'])
        if re.search(regex_file_number,final_filename.split('-')[8]):
            final_filename = final_filename.replace(final_filename.split('-')[8]+'-','') #remove the -004- giving the number of the file
    final_filename = final_filename.replace('l1b','l2')
    final_filename = final_filename.replace('l1c','l2')
    final_filename = final_filename.replace('xsp','wav')
    savepath = os.path.join(output_safe, final_filename)
    return savepath

def load_config():
    """

    Returns:
        conf: dict
    """
    local_config_path = os.path.join(os.path.dirname(sarwaveifrproc.__file__), 'localconfig.yaml')

    if os.path.exists(local_config_path):
        config_path = local_config_path
    else:
        config_path = os.path.join(os.path.dirname(sarwaveifrproc.__file__), 'config.yaml')
        
    logging.info('config path: %s', config_path)
    stream = open(config_path, 'r')
    conf = yaml.load(stream, Loader=yaml.CLoader)
    return conf


def load_models(paths, predicted_variables):
    """
    Loads models, scalers, and bins necessary for prediction.

    Parameters:
        paths (dict): Dictionary containing paths to model files, scaler files, and bin files.
            Keys:
                - 'model_intraburst': Path to the intraburst model file.
                - 'model_interburst': Path to the interburst model file.
                - 'scaler_intraburst': Path to the intraburst scaler file.
                - 'scaler_interburst': Path to the interburst scaler file.
                - 'bins_intraburst': Path to the intraburst bins directory.
                - 'bins_interburst': Path to the interburst bins directory.
        predicted_variables (list): List of variable names to be predicted.
    Returns:
        tuple: Tuple containing the following items:
            - model_intraburst (tf.keras.Model): Intraburst model loaded from the provided path.
            - model_interburst (tf.keras.Model): Interburst model loaded from the provided path.
            - scaler_intraburst (RobustScaler): Intraburst scaler loaded from the provided path.
            - scaler_interburst (RobustScaler): Interburst scaler loaded from the provided path.
            - bins_intraburst (dict): Dictionary containing intraburst bins for each predicted variable.
            - bins_interburst (dict): Dictionary containing interburst bins for each predicted variable.

    """    
    # Unpack paths
    path_model_intraburst, path_model_interburst, path_scaler_intraburst, path_scaler_interburst, path_bins_intraburst, path_bins_interburst = paths.values()
    
    # Load models and scalers using paths provided
    model_intraburst = tf.keras.models.load_model(path_model_intraburst, compile=False)
    scaler_intraburst_array = np.load(path_scaler_intraburst)
    scaler_intraburst = RobustScaler(medians=scaler_intraburst_array[0], iqrs=scaler_intraburst_array[1])    
    
    model_interburst = tf.keras.models.load_model(path_model_interburst, compile=False)
    scaler_interburst_array = np.load(path_scaler_interburst)
    scaler_interburst = RobustScaler(medians=scaler_interburst_array[0], iqrs=scaler_interburst_array[1]) 
    
    bins_intraburst = {f'{var}': np.load(os.path.join(path_bins_intraburst, f'bins_{var}.npy')) for var in predicted_variables}
    bins_interburst = {f'{var}': np.load(os.path.join(path_bins_interburst, f'bins_{var}.npy')) for var in predicted_variables}

    return model_intraburst, model_interburst, scaler_intraburst, scaler_interburst, bins_intraburst, bins_interburst
    
    
def process_files(input_safe, output_safe, model_intraburst, model_interburst, scaler_intraburst, scaler_interburst, bins_intraburst, bins_interburst, predicted_variables, product_id):
    """
    Processes files in the input directory, generates predictions, and saves results in the output directory.

    Parameters:
        input_safe (str): Input safe path.
        output_safe (str): Path to the directory where output data will be saved.
        model_intraburst (tf.keras.Model): Intraburst model for prediction.
        model_interburst (tf.keras.Model): Interburst model for prediction.
        scaler_intraburst (RobustScaler): Scaler for intraburst data.
        scaler_interburst (RobustScaler): Scaler for interburst data.
        bins_intraburst (dict): Dictionary containing intraburst bins for each predicted variable.
        bins_interburst (dict): Dictionary containing interburst bins for each predicted variable.
        predicted_variables (list): List of variable names to be predicted.
        product_id (str): Identifier for the output product.

    Returns:
        None
    """
    subswath_filenames = glob.glob(os.path.join(input_safe, '*-?v-*.nc'))
    logging.info(f'{len(subswath_filenames)} subswaths found in given safe.')
    
    for path in subswath_filenames:
        xdt = dtt.open_datatree(path)
        l2_product = generate_l2_wave_product(xdt, model_intraburst, model_interburst, scaler_intraburst, scaler_interburst, bins_intraburst, bins_interburst, predicted_variables)

        os.makedirs(output_safe, exist_ok=True)
        savepath = get_output_filename(path, output_safe, product_id)
        l2_product.to_netcdf(savepath)
        
    
class RobustScaler:
    """
    Class to mimic scikit-learn RobustScaler. This is done in order to prevent warning messages when using pickle to load the scikit-learn scaler.
    """

    def __init__(self, medians, iqrs):
        """
        Initialize the RobustScaler with provided medians and IQRs.

        Parameters:
            medians (np.ndarray): Median values for each feature.
            iqrs (np.ndarray): Interquartile ranges (IQRs) for each feature.
        """
        self.medians = medians
        self.iqrs = iqrs

    def transform(self, X):
        """
        Scale the input data X using the stored medians and IQRs.

        Parameters:
            X (np.ndarray): Input data to be scaled.

        Returns:
            X_scaled (np.ndarray): Scaled data.
        """
        X_scaled = (X - self.medians) / self.iqrs
        return X_scaled



import argparse
import logging
import os
import sys
import glob
import numpy as np
from sarwaveifrproc.utils import get_output_safe, load_config, load_models, process_files

def parse_args():
    
    parser = argparse.ArgumentParser(description="Generate a L2 WAVE product from a L1B or L1C SAFE.")

    # Define arguments
    parser.add_argument("--input_path", required=True, help="l1b or l1c safe path or listing path (.txt file).")
    parser.add_argument("--save_directory", required=True, help="where to save output data.")
    parser.add_argument("--product_id", required=True, help="3 digits ID representing the processing options. Ex: E00.")

    # Group related arguments under 'model' and 'bins'
    model_group = parser.add_argument_group('model', 'Arguments related to the neural models')
    model_group.add_argument("--model_intraburst", required=False, help="neural model path to predict sea states on intraburst data.")
    model_group.add_argument("--model_interburst", required=False, help="neural model path to predict sea states on interburst data.")

    model_group.add_argument("--scaler_intraburst", required=False, help="scaler path to standardize intraburst data before feeding it to the neural model.")
    model_group.add_argument("--scaler_interburst", required=False, help="scaler path to standardize interburst data before feeding it to the neural model.")

    model_group.add_argument("--bins_intraburst", required=False, help="bins path that depicts the range of predictions on intraburst data.")
    model_group.add_argument("--bins_interburst", required=False, help="bins path that depicts the range of predictions on interburst data.")

    model_group.add_argument("--predicted_variables", required=False, help="list of sea states variables to predict.")
    
    # Other arguments
    parser.add_argument("--overwrite", action="store_true", default=False, help="overwrite the existing outputs")
    parser.add_argument("--verbose", action="store_true", default=False)
   
    args = parser.parse_args()
    return args


def setup_logging(verbose=False):
    fmt = '%(asctime)s %(levelname)s %(filename)s(%(lineno)d) %(message)s'
    level = logging.DEBUG if verbose else logging.INFO
    logging.basicConfig(level=level, format=fmt, datefmt='%d/%m/%Y %H:%M:%S', force=True)

def get_files(dir_path, listing):
    
    fn = []
    for s in listing:
        search_path = os.path.join(dir_path, s.replace('WAVE', 'XSP_'), '*-?v-*.nc')
        fn+=glob.glob(search_path)

    print('Number of files :', len(fn))
    return fn


def main():
    args = parse_args()
    setup_logging(args.verbose)

    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'  # Suppress TensorFlow INFO and WARNING messages
    os.environ['CUDA_VISIBLE_DEVICES'] = '-1'  # Hide CUDA devices

    logging.info("Loading configuration file...")
    conf = load_config()

    input_path = args.input_path
    save_directory = args.save_directory
    product_id = args.product_id
    predicted_variables = args.predicted_variables or conf['predicted_variables']

    if input_path.endswith('.txt'):
        files = np.loadtxt(input_path, dtype=str)
        output_safes = np.array([get_output_safe(f, save_directory, product_id) for f in files])

        if not args.overwrite:
            mask = np.array([os.path.exists(f) for f in output_safes])
            files, output_safes = files[~mask], output_safes[~mask]

            logging.info(f"{np.sum(mask)} file(s) already exist(s) and overwriting is not allowed. Use --overwrite to overwrite existing files.")

            if not files.size:
                return None

        logging.info('Loading models...')
        # Define the paths. When args paths are None, the conf paths are set by default (the "or" operator returns first element if true, second element if first is None). 
        paths = {
        'model_intraburst': args.model_intraburst or conf['model_intraburst'],
        'model_interburst': args.model_interburst or conf['model_interburst'],
        'scaler_intraburst': args.scaler_intraburst or conf['scaler_intraburst'],
        'scaler_interburst': args.scaler_interburst or conf['scaler_interburst'],
        'bins_intraburst': args.bins_intraburst or conf['bins_intraburst'],
        'bins_interburst': args.bins_interburst or conf['bins_interburst'],
        }
        model_intraburst, model_interburst, scaler_intraburst, scaler_interburst, bins_intraburst, bins_interburst = load_models(paths, predicted_variables)
        logging.info('Models loaded.')

        logging.info('Processing files...')
        for f, output_safe in zip(files, output_safes):
            process_files(f, output_safe, model_intraburst, model_interburst, scaler_intraburst, scaler_interburst, bins_intraburst, bins_interburst, predicted_variables, product_id)

            
    else:
        logging.info("Checking if output safe already exists...")
        output_safe = get_output_safe(input_path, save_directory, product_id)

        if os.path.exists(output_safe) and not args.overwrite:
            logging.info(f"{output_safe} already exists and overwriting is not allowed. Use --overwrite to overwrite existing files.")
            return None

        logging.info('Loading models...')
        paths = {
            'model_intraburst': args.model_intraburst or conf['model_intraburst'],
            'model_interburst': args.model_interburst or conf['model_interburst'],
            'scaler_intraburst': args.scaler_intraburst or conf['scaler_intraburst'],
            'scaler_interburst': args.scaler_interburst or conf['scaler_interburst'],
            'bins_intraburst': args.bins_intraburst or conf['bins_intraburst'],
            'bins_interburst': args.bins_interburst or conf['bins_interburst'],
        }
        model_intraburst, model_interburst, scaler_intraburst, scaler_interburst, bins_intraburst, bins_interburst = load_models(paths, predicted_variables)
        logging.info('Models loaded.')

        logging.info('Processing files...')
        process_files(input_path, output_safe, model_intraburst, model_interburst, scaler_intraburst, scaler_interburst, bins_intraburst, bins_interburst, predicted_variables, product_id)

    logging.info(f'Processing terminated. Output directory: \n{save_directory}')



import numpy as np
import pytest
import xarray as xr
import os
import logging
import sarwaveifrproc
# reference prediction is E01 (rmarquart January 2024), validated by L.Maillard
def get_hs_values(ff):
    ds=xr.open_dataset(ff,group='intraburst')
    return ds['hs_mean'].data
subswathSAR=[os.path.join(os.path.dirname(sarwaveifrproc.__file__),'reference_data',
                          's1a-iw2-slc-dv-20231128t035702-20231128t035727-051412-063451-e01.nc'),
                       ]
# L1B_SAR_vh = [os.path.join(os.path.dirname(sarwaveifrproc.__file__),'reference_data','s1a-iw2-slc-vh-20231128t035702-20231128t035727-051412-063451-002_L1B_xspec_IFR_3.7.6nospectra.nc')]
L1B_SAR_vv = [os.path.join(os.path.dirname(sarwaveifrproc.__file__),'reference_data',
            's1a-iw2-slc-vv-20231128t035702-20231128t035727-051412-063451-005_L1B_xspec_IFR_3.7.6nospectra.nc')]
hs_expected = []
for ii in subswathSAR:
    # hs_expected[ii] = get_hs_values(ff=ii)
    hs_expected.append(get_hs_values(ff=ii))

@pytest.mark.parametrize(
    "L1B_SAR_vv, hs_expected",
    [(L1B_SAR_vv[0],hs_expected[0])],
)
def test_hs_prediction(L1B_SAR_vv,hs_expected):
    from sarwaveifrproc.utils import  load_config,load_models
    ds = xr.open_dataset(L1B_SAR_vv) # no group in this light l1B without interburst and without xspectra
    conf = load_config()
    predicted_variables = conf['predicted_variables']
    paths = {
        'model_intraburst': conf['model_intraburst'],
        'model_interburst': conf['model_interburst'],
        'scaler_intraburst': conf['scaler_intraburst'],
        'scaler_interburst': conf['scaler_interburst'],
        'bins_intraburst': conf['bins_intraburst'],
        'bins_interburst': conf['bins_interburst'],
    }
    logging.info('Loading models...')
    (model_intraburst, model_interburst, scaler_intraburst, scaler_interburst,
     bins_intraburst, bins_interburst) = load_models(
        paths, predicted_variables)
    logging.info('Models loaded.')
    # kept_variables = ['corner_longitude', 'corner_latitude', 'land_flag', 'sigma0_filt', 'normalized_variance_filt',
    #                   'incidence', 'azimuth_cutoff', 'cwave_params']
    kept_variables = []
    dspredicted = sarwaveifrproc.l2_wave.generate_intermediate_product(ds,
                                                model_intraburst, scaler_intraburst,
                                                bins_intraburst,
                                                predicted_variables, kept_variables)
    hs_actual = dspredicted['hs_mean'].data
    logging.info('hs_actual %s',hs_actual)
    logging.info('hs_expected %s', hs_expected)
    mask_nan = np.isfinite(hs_expected)
    diff = hs_actual[mask_nan]-hs_expected[mask_nan]
    logging.info('diff : %s %s',diff.min(),diff.max())
    assert np.allclose(hs_actual[mask_nan],hs_expected[mask_nan],atol=0.1)


if __name__ == '__main__':
    import logging
    logging.basicConfig(level=logging.INFO)
    logging.info('start')
    test_hs_prediction(L1B_SAR_vv[0],hs_expected[0])



import numpy as np
import os
import xarray as xr
import datatree as dtt
from scipy import special
import sarwaveifrproc

attributes_missing_variables = \
{
    'sigma0_filt': {'long_name': 'calibrated sigma0 with BT correction', 'units': 'linear'},
    'normalized_variance_filt': {'long_name': 'normalized variance with BT correction', 'units': ''},
    'azimuth_cutoff': {'long_name': 'Azimuthal cut-off (2tau)', 'units': 'm'},
    'cwave_params': {'long_name': 'CWAVE parameters'},
    'k_gp': {'long_name': 'Gegenbauer polynoms dimension'},
    'phi_hf': {'long_name': 'Harmonic functions dimension (odd number)'}
}

def generate_l2_wave_product(xdt, intraburst_model, interburst_model, intraburst_scaler, interburst_scaler, intraburst_bins, interburst_bins, predicted_variables):
    """
    Generate a level-2 wave (L2 WAV) product.

    Parameters:
    - xdt (dict): DataTree containing intraburst and interburst datasets.
    - intraburst_model (tf.keras.Model): Machine learning model for intraburst predictions.
    - interburst_model (tf.keras.Model): Machine learning model for interburst predictions.
    - intraburst_scaler (Union[StandardScaler, MinMaxScaler, RobustScaler]): Scaler used for intraburst normalization.
    - interburst_scaler (Union[StandardScaler, MinMaxScaler, RobustScaler]): Scaler used for interburst normalization.
    - intraburst_bins (dict): Dictionary containing bins for intraburst variables.
    - interburst_bins (dict): Dictionary containing bins for interburst variables.
    - predicted_variables (list): List of predicted variable names.
    Returns:
    - l2_wave_product (dtt.DataTree): Level-2 wave product.
    
    Notes:
    - The scaler objects should be one of StandardScaler, MinMaxScaler, or RobustScaler from sklearn.preprocessing.
    """

    kept_variables = ['corner_longitude', 'corner_latitude', 'land_flag', 'sigma0_filt', 'normalized_variance_filt', 'incidence', 'azimuth_cutoff', 'cwave_params']
    ds_intraburst = generate_intermediate_product(xdt['intraburst'].ds, intraburst_model, intraburst_scaler, intraburst_bins, predicted_variables, kept_variables)
    ds_interburst = generate_intermediate_product(xdt['interburst'].ds, interburst_model, interburst_scaler, interburst_bins, predicted_variables, kept_variables)
    
    l2_wave_product = dtt.DataTree.from_dict({"intraburst": ds_intraburst, "interburst": ds_interburst})
    l2_wave_product.attrs['input_sar_product'] = os.path.basename(xdt.encoding['source'])
    return l2_wave_product


def generate_intermediate_product(ds, model, scaler, bins, predicted_variables, kept_variables):
    """
    Generate an intermediate l2 product, depending of the input dataset (intraburst or interburst).

    Parameters:
    - ds (xarray.Dataset): Input dataset.
    - model: Machine learning model for predicting sea parameters.
    - scaler: Scaler used for normalization.
    - bins (dict): Dictionary containing bins for each variable.
    - predicted_variables (list): List of predicted variable names.
    - kept_variables (list): List of variables from the input dataset that are kept in the final product.

    Returns:
    - ds_pred (xarray.Dataset): Intermediate predictions dataset.
    """
    
    if ds['land_flag'].all():
        return generate_product_on_land(ds, predicted_variables, kept_variables, bins)
    
    if '2tau' in ds.dims:
        ds = ds.squeeze(dim='2tau')
        ds.attrs['squeezed_dimensions']='2tau'
    
    tiles = ds[['sigma0_filt', 'normalized_variance_filt',  'incidence', 'azimuth_cutoff', 'cwave_params']]
    if 'burst' in ds.coords:
        tiles_stacked = tiles.stack(all_tiles = ['burst', 'tile_line','tile_sample'], k_phi = ['phi_hf', 'k_gp'])
    else:
        tiles_stacked = tiles.stack(all_tiles = ['tile_line','tile_sample'], k_phi = ['phi_hf', 'k_gp'])
            
    output_dims = [['all_tiles', f'{v}_mid'] for v in predicted_variables]
    inds = np.cumsum([0] + [len(bins[v]) -1 for v in predicted_variables])
    
    predictions = xr.apply_ufunc(predict_variables,
                                 tiles_stacked['sigma0_filt'], tiles_stacked['normalized_variance_filt'], tiles_stacked['incidence'], 
                                 tiles_stacked['azimuth_cutoff'], tiles_stacked['cwave_params'],
                                 model, scaler, inds,
                                 input_core_dims=[['all_tiles'], ['all_tiles'], ['all_tiles'], ['all_tiles'], ['all_tiles','k_phi'], [], [], []],
                                 output_core_dims=output_dims,
                                 vectorize=False)
    
    ds_pred = format_dataset(ds, predictions, predicted_variables, kept_variables, bins)
    
    return ds_pred

def generate_product_on_land(ds, predicted_variables, kept_variables, bins):
    """ 
    Patch function when input dataset does not contain all necessary variables.
    
    Parameters:
    - ds (xarray.Dataset): Input dataset.
    - predicted_variables (list): List of predicted variable names.
    - kept_variables (list): List of variables from the input dataset that are kept in the final product.
    - bins (dict): Dictionary containing bins for each variable.
    """
    ds = ds[['corner_longitude', 'corner_latitude', 'incidence', 'land_flag']]
    
    shape = ds['land_flag'].shape
    dims = ds['land_flag'].dims
    
    data_to_merge = []

    if set(predicted_variables).issubset(ds.keys()):
        kept_variables = kept_variables + predicted_variables
    
    for v in kept_variables:
        if v not in ds.variables:
            v_array = xr.DataArray(data = np.full(shape, np.nan), dims = dims).rename(v)
            v_array.attrs = attributes_missing_variables[v]
            data_to_merge.append(v_array)
            
    # Add CWAVES independently because the required dimensions are not in the input dataset when there is only land.
    k_gp = xr.DataArray(data=range(1, 5), dims='k_gp')
    k_gp.attrs = attributes_missing_variables['k_gp']
    phi_hf = xr.DataArray(data=range(1, 6), dims='phi_hf')
    phi_hf.attrs = attributes_missing_variables['phi_hf']
    
    cwaves = xr.DataArray(data=np.full(shape + (4, 5), np.nan), dims=dims + ('k_gp', 'phi_hf')).rename('cwave_params')
    cwaves.attrs = attributes_missing_variables['cwave_params']
    data_to_merge.append(cwaves.assign_coords({'k_gp': k_gp, 'phi_hf': phi_hf}))
    
    for v in predicted_variables:
        attributes = get_attributes(v)
        v_mid = xr.DataArray(data=(bins[v][:-1] + bins[v][1:])/2, dims=f'{v}_mid')
        v_mid.attrs = attributes['mid']
        
        v_pdf = xr.DataArray(data=np.full(v_mid.shape + shape, np.nan), dims=(f'{v}_mid', ) + dims).rename(f'{v}_pdf')
        v_pdf.attrs = attributes['pdf']
        data_to_merge.append(v_pdf.assign_coords({f'{v}_mid': v_mid}))
                
        v_mean = xr.DataArray(data=np.full(shape, np.nan), dims=dims).rename(f'{v}_mean')
        v_mean.attrs = attributes['mean']
        data_to_merge.append(v_mean)

        v_most_likely = xr.DataArray(data=np.full(shape, np.nan), dims=dims).rename(f'{v}_most_likely') 
        v_most_likely.attrs = attributes['most_likely']
        data_to_merge.append(v_most_likely)

        v_std = xr.DataArray(data=np.full(shape, np.nan), dims=dims).rename(f'{v}_std') 
        v_std.attrs = attributes['std']
        data_to_merge.append(v_std)
        
    ds = xr.merge([ds] + data_to_merge)                         
    
    ds.attrs.pop('name', None)
    ds.attrs.pop('multidataset', None)
    return ds

def predict_variables(sigma0, normalized_variance, incidence, azimuth_cutoff, cwave_params, model, scaler, indices):
    """
    Launch predictions using a neural model.

    Parameters:
    - sigma0 (xarray.DataArray): Array containing sigma0 values.
    - normalized_variance (xarray.DataArray): Array containing normalized variance values.
    - incidence (xarray.DataArray): Array containing incidence values.
    - azimuth_cutoff (xarray.DataArray): Array containing azimuth cutoff values.
    - cwave_params (xarray.DataArray): Array containing the cwave parameters.
    - model (tf.keras.Model): Machine learning model for predicting sea parameters.
    - scaler (Union[StandardScaler, MinMaxScaler, RobustScaler]): Scaler used for normalization.
    - indices (list): List of indices to segment predictions.

    Returns:
    - res (tuple): Tuple containing predictions for each variable.
    """   
    X_stacked = np.vstack([sigma0, normalized_variance, incidence, azimuth_cutoff]).T
    X_stacked = np.hstack([X_stacked, cwave_params])
    X_normalized = scaler.transform(X_stacked)
    
    predictions = model.predict(X_normalized)
    
    res = tuple(predictions[:, indices[i]:indices[i+1]] for i in range(len(indices)-1))
    
    return res

def format_dataset(ds, predictions, predicted_variables, kept_variables, bins):
    """
    Format a dataset based on predictions, variables, and bins.

    Parameters:
    - ds (xarray.Dataset): Input dataset (l1b or l1c).
    - predictions (list): Predictions for each variable.
    - predicted_variables (list): List of predicted variable names.
    - kept_variables (list): List of variables from the input dataset that are kept in the final product.
    - bins (dict): Dictionary containing bin values for each variable.

    Returns:
    - ds_pred (xarray.Dataset): Formatted dataset with predictions and original data.

    The function computes various statistical values (mean, most likely, standard deviation)
    from the predictions and adds them to the new dataset. It also includes original data
    from the input dataset and some additional attributes.
    """
        
    data_to_merge = []
    
    n = len(predicted_variables)
    for i in range(n):
        v = predicted_variables[i]
        attributes = get_attributes(v)
        v_mid = xr.DataArray(data = (bins[v][:-1] + bins[v][1:])/2, dims = f'{v}_mid')
        v_mid.attrs = attributes['mid']

        v_pdf = xr.apply_ufunc(lambda x: special.softmax(x, axis=1),
                               predictions[i]).rename(f'{v}_pdf').assign_coords({f'{v}_mid': v_mid}).to_dataset()
        v_pdf[f'{v}_pdf'].attrs = attributes['pdf']
        data_to_merge.append(v_pdf)
        
        v_mean = compute_values(v_pdf, v, compute_mean).rename(f'{v}_mean')
        v_mean.attrs = attributes['mean']
        data_to_merge.append(v_mean)
        
        v_most_likely = compute_values(v_pdf, v, get_most_likely).rename(f'{v}_most_likely')
        v_most_likely.attrs = attributes['most_likely']
        data_to_merge.append(v_most_likely)
        
        v_std = compute_values(v_pdf, v, compute_std, True).rename(f'{v}_std')
        v_std.attrs = attributes['std']
        data_to_merge.append(v_std)

    if set(predicted_variables).issubset(ds.keys()):
        kept_variables = kept_variables + predicted_variables
            
    data_to_merge = [ds.unstack() for ds in data_to_merge]
    ds = ds[kept_variables]
    
    ds = xr.merge([ds] + data_to_merge)
    ds = ds.drop_vars(['tile_line', 'tile_sample'])        
    ds.attrs['l2_processor_name'] = 'sarwaveifrproc'
    ds.attrs['l2_processor_version'] = sarwaveifrproc.__version__
    ds.attrs.pop('name', None)
    ds.attrs.pop('multidataset', None)
    
    return ds

def compute_values(ds, var, function, vectorize=False):
    """
    Compute values for the given variable using the given function.

    Args:
        ds (xr.Dataset): Dataset containing the data.
        var (str): Variable name.
        function (callable): Function to compute values.
        vectorize (bool): Whether to vectorize the computation.

    Returns:
        xr.DataArray: Computed values.
    """
    values = xr.apply_ufunc(function,
                            ds[f'{var}_mid'], ds[f'{var}_pdf'], 
                            input_core_dims=[[f'{var}_mid'],[f'{var}_mid']],
                            vectorize=vectorize)
    return values 

def get_attributes(var):
    """
    Generate a dictionary of attributes for a given variable.

    Parameters:
    - var (str): The variable name for which attributes are to be generated. This should correspond to keys in the `spec_dict`.

    Returns:
    - attributes (dict): A dictionary containing various attributes for the given variable.

    The function uses a predefined `spec_dict` to look up information about the variable.
    If the variable is not found in `spec_dict`, default attributes are generated with the variable name itself.
    """
    spec_dict = {
           'hs': {'long_name': 'significant wave height', 'units': 'm'},
           'phs0': {'long_name': 'wind sea significant wave height', 'units': 'm'},
           't0m1': {'long_name': 'mean wave period', 'units': 's'},
            }
    
    var_spec = spec_dict.get(var, {"long_name": f'{var}', 'units': ''})
    
    attributes = {
     'mid': {'long_name': f'central values of the bins used for discretizing the range of {var_spec["long_name"]} encountered during neural model training', 'units': var_spec['units']},
     'pdf': {'long name': f'{var_spec["long_name"]} discrete probability density function', 'units': 'probability'},
     'mean': {'long name': f'first-order moment of the {var_spec["long_name"]} discrete probability density function', 'units': var_spec['units']},
     'most_likely': {'long name': f'most likely value of {var_spec["long_name"]} given its discrete probability density function', 'units': var_spec['units']},
     'std': {'long name': f'square root of the second-order moment of the {var_spec["long_name"]} discrete probability density function', 'units': var_spec['units']}
    }
    
    return attributes

def get_most_likely(x, y):
    """
    Get the maximum of probability for each prediction.

    Args:
        x (np.ndarray): Input values.
        y (np.ndarray): Probabilities.

    Returns:
        np.ndarray: Most likely values.
    """
    i_max = np.argmax(y, axis=1)
    most_likely = x[i_max]
    return most_likely

def compute_mean(x, y):
    """
    Compute the expected value.

    Args:
        x (np.ndarray): Input values.
        y (np.ndarray): Probabilities.

    Returns:
        np.ndarray: Expected value.
    """
    return np.sum(x * y, axis=1)

def compute_std(x, y):
    """
    Compute the standard deviation.

    Args:
        x (np.ndarray): Input values.
        y (np.ndarray): Probabilities.

    Returns:
        np.ndarray: Standard deviation.
    """
    mean = np.sum(x * y)
    variance = np.sum(y * (x - mean) ** 2)
    return np.sqrt(variance)
